;; Generated package description from objed.el
(define-package "objed" "0.3.1" "Navigate and edit text objects." '((emacs "25") (cl-lib "0.5")) :url "https://github.com/clemera/objed" :keywords '("convenience"))
